#include "glutwm.h"
#include <glui.h>

#include "error.h"
#include "cloud.h"
#include "TriMesh.h"
#include "samplewindow.h"
#include "trimeshutil.h"
#include "pointsample.h"
#include "buffer.h"
#include "time.h"
#include "tribox.h"

TriMesh *mesh;
Cloud cloud;
Array<char> mask;
SamplerOctree *tree;
SampleWindow wsample;

static struct {
    int level;
    float lambda;
    float bad;
    int nsamples;
} glui;

static void stratified_cb(int)
{
    delete tree;
    tree = StratifiedSample(mesh, 
            glui.level, glui.lambda, glui.bad, 
            &cloud, mask);
    wsample.SetOctree(tree);
    wsample.PostRedisplay();
}

static void point_cb(int)
{
    PointSample(mesh, glui.nsamples, &cloud);
    mask.Resize(cloud.vertex.GetSize());
    for (int i = 0; i < mask.GetSize(); i++) mask[i] = 1;
    wsample.SetOctree(NULL);
    wsample.PostRedisplay();
}

static void dummy_cb(int)
{
}

static void setidle(void (*func)(void))
{
    GLUI_Master.set_glutIdleFunc(func);
}

static void InitGLUI(void)
{
    GLUI *control = GLUI_Master.create_glui("Control");
    // Sampling pannel
    GLUI_Panel *panel = control->add_panel("Uniform");
    GLUI_Spinner *spinner = control->add_spinner_to_panel(panel,
            "Samples", GLUI_SPINNER_INT, &glui.nsamples, 
            0, dummy_cb);
    spinner->set_int_limits(10, 1000000, GLUI_LIMIT_CLAMP);
    glui.nsamples = 1000;
    control->add_button_to_panel(panel, "Sample", 0, point_cb);

    panel = control->add_panel("Stratified");
    spinner = control->add_spinner_to_panel(panel,
            "Level", GLUI_SPINNER_INT, &glui.level, 
            0, dummy_cb);
    spinner->set_int_limits(1, 10, GLUI_LIMIT_CLAMP);
    glui.level = 6;
    spinner = control->add_spinner_to_panel(panel,
            "Lambda", GLUI_SPINNER_FLOAT, &glui.lambda, 
            0, dummy_cb);
    spinner->set_float_limits(0.001, 1000, GLUI_LIMIT_CLAMP);
    glui.lambda = 10;
    spinner = control->add_spinner_to_panel(panel,
            "Bad", GLUI_SPINNER_FLOAT, &glui.bad, 
            0, dummy_cb);
    spinner->set_float_limits(0.001, 5, GLUI_LIMIT_CLAMP);
    glui.bad = 0.5;
    control->add_button_to_panel(panel, "Sample", 0, stratified_cb);

	GLUI_Master.sync_live_all();
	GlutWMMetaIdleFunc(setidle);
}

int main(int argc, char *argv[])
{
    GlutWMOpen(&argc, argv);
    GlutWMInitWindowSize(512, 512);
    if (argc < 2) {
        fprintf(stderr, "Usage\n  sample <ply1> [.. <plyN>]\n");
        return 1;
    }
    srand48(clock());
    debugf(("Loading mesh"));
    mesh = TriMesh::read(argv[1]);
    if (!mesh) {
        errorf(("Unable to open file '%s'", argv[1]));
        return 1;
    }
    debugf(("%d vertices, %d faces", mesh->vertices.size(), 
                mesh->faces.size()));
//    Normalize(mesh);

    wsample.SetCloud(&cloud);
    wsample.SetMask(&mask);
    wsample.SetTriMesh(mesh);
    wsample.Fit();
    GlutWMCreateWindow("Sampling viewer", &wsample);

	InitGLUI();

    GlutWMMainLoop();
    return 0;
}
