#include <glui.h>
#include <algorithm>

#include "modelwindow.h"
#include "array.h"
#include "matrix.h"
#include "sample.h"
#include "pngimage.h"

void drawcircle(double x, double y, double r)
{
    glBegin(GL_LINE_LOOP);
    for (double a = 0; a < 2*SCALAR_PI; a += SCALAR_PI/32) {
        glVertex3d(x + r*cos(a), y + r*sin(a), 0);
    }
    glEnd();
}

void drawbox(const Vector3 &min, const Vector3 &max)
{
    glBegin(GL_LINE_LOOP);
    glVertex3d(min[0], min[1], min[2]);
    glVertex3d(max[0], min[1], min[2]);
    glVertex3d(max[0], max[1], min[2]);
    glVertex3d(min[0], max[1], min[2]);
    glEnd();
    glBegin(GL_LINE_LOOP);
    glVertex3d(min[0], min[1], max[2]);
    glVertex3d(max[0], min[1], max[2]);
    glVertex3d(max[0], max[1], max[2]);
    glVertex3d(min[0], max[1], max[2]);
    glEnd();
    glBegin(GL_LINES);
    glVertex3d(min[0], min[1], min[2]);
    glVertex3d(min[0], min[1], max[2]);
    glVertex3d(max[0], min[1], min[2]);
    glVertex3d(max[0], min[1], max[2]);
    glVertex3d(max[0], max[1], min[2]);
    glVertex3d(max[0], max[1], max[2]);
    glVertex3d(min[0], max[1], min[2]);
    glVertex3d(min[0], max[1], max[2]);
    glEnd();
}

void drawpoint(Vector3 p, double r)
{
    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glTranslated(p[0], p[1], p[2]);
    glutSolidSphere(r/10, 16, 16);
    glPopMatrix();
}

static int color2i(GLubyte *color)
{
    int i = color[2];
    i <<= 8;
    i |= color[1];
    i <<= 8;
    i |= color[0];
    return (i == 0xffffff) ? -1: i;
}

static void i2color(int i, GLubyte *color)
{
    color[0] = (GLubyte) (i & 0xff);
    i >>= 8;
    color[1] = (GLubyte) (i & 0xff);
    i >>= 8;
    color[2] = (GLubyte) (i & 0xff);
}

class SampleWindow: public ModelWindow {
public:

    SampleWindow(void)
    {
        control[0] = Vector3(1, 0, 0);
        control[1] = Vector3(0, 1, 0);
        control[2] = Vector3(-1, 0, 0);
		control[3] = Vector3(0, -1, 0);
		control[4] = Vector3(-0.5, -0.5, -0.1);
		control[5] = Vector3(0.5, 0.5, 0.1);
        box.Merge(control[4]);
        box.Merge(control[5]);
        selected = -1;
        mode = "exponential";
        nsamples = 100;
        lambda = 20;
        radius = 1;
    }

    void SetMode(const char *_mode)
    {
        if (!strcmp(_mode, "uniform")) mode = "uniform";
        else if (!strcmp(_mode, "exponential")) mode = "exponential";
    }

    void ColorControl(void)
    {
        glClearColor(1, 1, 1, 1);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        glEnable(GL_DEPTH_TEST);
        glDepthFunc(GL_LESS);
        glDisable(GL_LIGHTING);
        glDisable(GL_BLEND);
        glPointSize(10);
        glBegin(GL_POINTS);
        // loop through each vertex drawing it in a unique color
        for (int i = 0; i < 6; i++) {
            GLubyte color[3];
            i2color(i, color);
            glColor3ubv(color);
            glVertex3fv(&control[i][0]);
        }
        glEnd();
    }

    int PickByColor(int mx, int my)
    {
        ColorControl();
        GLubyte color[3];
        glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
        GlutWMReadPixels(mx, my, 1, 1, GL_RGB, color);
        return color2i(color);
    }

    void DrawModel(void)
    {
		glDisable(GL_LIGHTING);
        glDisable(GL_DEPTH_TEST);
		glClearColor(1, 1, 1, 1);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        glColor3d(0, 0, 0);
        glLineWidth(2);
        glBegin(GL_LINE_LOOP);
            glVertex3fv(&control[0][0]);
            glVertex3fv(&control[1][0]);
            glVertex3fv(&control[2][0]);
        glEnd();
		// draw selected
        if (selected >= 0) {
            glPointSize(10);
            glColor3d(1, 0, 0);
            drawpoint(control[selected], radius);
            Vector3 c = ClosestToPointInTriangle(control[3], control[0], 
                    control[1], control[2]);
            drawpoint(c, radius);
            drawcircle(control[3][0], control[3][1], Norm(c - control[3]));
            // c = FurthestToPointInTriangle(control[3], control[0], 
              //       control[1], control[2]);
            // drawcircle(control[3][0], control[3][1], Norm(c - control[3]));
        }
		// draw samples
		//Sample(nsamples, lambda);
        glPointSize(3);
        glColor3d(0, 0.5, 0.5);
        for (int i = 0; i < sample.GetSize(); i++)
            drawpoint(sample[i], radius);
        glPointSize(10);
		// draw control points
        glColor3d(0, 0, 1);
        for (int i = 0; i < 6; i++)
            drawpoint(control[i], radius);
		// draw box
        glColor3d(1, 0.5, 0.5);
        drawbox(box.GetMin(), box.GetMax());
    }

    void Fit(void)
    {
        SetCenter(Vector3());
        SetSize(3);
    }

	void Button(int button, int state, int x, int y)
	{
        if (!(glutGetModifiers() & GLUT_ACTIVE_CTRL)) {
            if (button == GLUT_LEFT_BUTTON) {
                if (state == GLUT_DOWN) {
                    selected = PickByColor(x, y);
                    drag = true;
                } else drag = false;
                PostRedisplay();
            } 
        } else ModelWindow::Button(button, state, x, y);
    }

	void Screen2World(double *x, double *y)
	{
		int w = GetWidth();
		int h = GetHeight();
		*x = 3.0*(*x - w/2)/w;
		*y = 3.0*(*y - h/2)/h;
	}

	void Motion(int x, int y)
	{
		double wx = x;
		double wy = y;
		Screen2World(&wx, &wy);
        if (drag && selected >= 0)
            control[selected] = Vector3(wx, wy, 0);
		ModelWindow::Motion(x, y);
        PostRedisplay();
        box.Reset();
        box.Merge(control[4]);
        box.Merge(control[5]);
	}

    void Sample(int _nsamples, double _lambda)
    {
		nsamples = _nsamples;
        lambda = _lambda;
        sample.Resize(0);
        switch (mode[0]) {
            case 'u':
                for (int i = 0; i < nsamples; i ++)
                    sample.PushBack(UniformPoint(control[0], 
                                control[1], control[2]));
                break;
            case 'e':
                for (int i = 0; i < nsamples; i ++) {
                    Vector3 p;
                    ExponentialPoint(control[3], lambda, 
                                Norm(box.GetSize())/2, control[0], control[1], 
                                control[2], box, &p);
                    sample.PushBack(p);
                }
                break;
        }
    }

    void Keyboard(unsigned char c, int x, int y)
    {
        switch (c) {
            case ',': 
                radius /= 2;
                PostRedisplay();
                break;
            case '.':
                radius *= 2;
                PostRedisplay();
                break;
            default:
                ModelWindow::Keyboard(c, x, y);
        } 
    }

private:
    Vector3 control[6];
    Array<Vector3> sample;
    BBox3 box;
    bool drag;
    double radius;
    const char *mode;
	int nsamples, selected;
    double lambda;
};

SampleWindow wsample;

static struct {
    int nsamples;
    float lambda;
    int mode;
} glui;

static void sample_cb(int)
{
    wsample.Sample(glui.nsamples, glui.lambda);
    wsample.PostRedisplay();
}

static void mode_cb(int)
{
    if (glui.mode == 0) wsample.SetMode("exponential");
    else wsample.SetMode("uniform"); 
    sample_cb(0);
}


static void setidle(void (*func)(void))
{
    GLUI_Master.set_glutIdleFunc(func);
}

static void InitGLUI(void)
{
    GLUI *control = GLUI_Master.create_glui("Control");
    // Sampling pannel
    GLUI_Panel *panel = control->add_panel("Sampling");
    GLUI_Spinner *spinner = control->add_spinner_to_panel(panel,
            "Samples", GLUI_SPINNER_INT, &glui.nsamples, 
            0, sample_cb);
    spinner->set_int_limits(0, 100000, GLUI_LIMIT_CLAMP);
    glui.nsamples = 100;

    spinner = control->add_spinner_to_panel(panel,
            "Lambda", GLUI_SPINNER_FLOAT, &glui.lambda, 
            0, sample_cb);
    spinner->set_float_limits(1, 100, GLUI_LIMIT_CLAMP);
    glui.lambda = 20;

    GLUI_RadioGroup *group = control->add_radiogroup_to_panel(panel,
            &glui.mode, 0, mode_cb);
    control->add_radiobutton_to_group(group, "Exponential");
    control->add_radiobutton_to_group(group, "Uniform");

    control->add_button_to_panel(panel, "Sample", 0, sample_cb);

	GLUI_Master.sync_live_all();
	GlutWMMetaIdleFunc(setidle);
}

int main(int argc, char *argv[])
{
    GlutWMOpen(&argc, argv);
    GlutWMInitWindowSize(480, 480);
    GlutWMCreateWindow("Sampling viewer", &wsample);
	wsample.Fit();
	InitGLUI();
    GlutWMMainLoop();
    return 0;
}
