#include <cstdio>
#include <cstdlib>
#include <algorithm>

#include "glutwm.h"
#include "trimeshwindow.h"

TriMeshWindow::
~TriMeshWindow()
{
    ;
}

void 
TriMeshWindow::
Clear(void) {
	glClearColor(0, 0, 0, 1);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}


void
TriMeshWindow::
SetupUniforms(void) {
    // setup basic environment
    glEnable(GL_NORMALIZE);
    glDisable(GL_CULL_FACE);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LESS);
    glEnable(GL_COLOR_MATERIAL);
    GLfloat s[] = {.3, .3, .3, 1};
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, s);
    glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 128);
    glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);
    glColor3f(1, 1, 1);
    glShadeModel(dflat? GL_FLAT : GL_SMOOTH);
}

void
TriMeshWindow::
DrawModel(void) {
    // setup vertex array stuff
    glEnableClientState(GL_VERTEX_ARRAY);
    glVertexPointer(3, GL_FLOAT, sizeof(mesh->vertices[0]), 
            &mesh->vertices[0][0]);
     glEnableClientState(GL_NORMAL_ARRAY);
     glNormalPointer(GL_FLOAT, sizeof(mesh->normals[0]), 
           &mesh->normals[0][0]);
    if (!mesh->colors.empty()) {
        glEnableClientState(GL_COLOR_ARRAY);
        glColorPointer(3, GL_UNSIGNED_BYTE, sizeof(mesh->colors[0]), 
                    &mesh->colors[0][0]);
        glEnable(GL_COLOR_MATERIAL);
    }
    // draw trimesh
    if (!dvertices) {
        if (!dflat) {
            glEnable(GL_LIGHTING);
            const int *t = &mesh->tstrips[0];
            const int *end = t + mesh->tstrips.size();
            while (t < end) {
                glDrawElements(GL_TRIANGLE_STRIP, *t, GL_UNSIGNED_INT, t+1);
                t += 1 + *t;
            }
        } else {
            glEnable(GL_LIGHTING);
            glBegin(GL_TRIANGLES);
            for (size_t i = 0; i < mesh->faces.size(); i++) {
                const Vector3 &p0(&mesh->vertices[mesh->faces[i][0]][0]);
                const Vector3 &p1(&mesh->vertices[mesh->faces[i][1]][0]);
                const Vector3 &p2(&mesh->vertices[mesh->faces[i][2]][0]);
                Vector3 n = Cross(p1 - p0, p2 - p0);
                n = Normalize(n);
                glNormal3fv(&n[0]);
                glVertex3fv(&p0[0]);
                glVertex3fv(&p1[0]);
                glVertex3fv(&p2[0]);
            }
            glEnd();
        }
    } else {
        glColor3f(0, 0, 0);
        glPointSize(3);
        glDisable(GL_LIGHTING);
        glDrawArrays(GL_POINTS, 0, mesh->vertices.size());
    }
    // draw mesh normals
    if (dnormals) {
        glDisable(GL_LIGHTING);
        glColor3f(0, 0, 1);
        glBegin(GL_LINES);
        for (size_t i = 0; i < mesh->vertices.size(); i++) {
            glVertex3fv(mesh->vertices[i]);
            glVertex3fv(mesh->vertices[i] + feature * mesh->normals[i]);
        }
        glEnd();
    }
}

void
TriMeshWindow::
SetTriMesh(TriMesh *_mesh)
{
    mesh = _mesh;
    if (mesh) {
        mesh->need_normals();
        mesh->need_tstrips();
        mesh->need_bsphere();
        mesh->need_faces();
    }
}

TriMesh * 
TriMeshWindow::
GetTriMesh(void)
{
    return mesh;
}


void
TriMeshWindow::
Fit(void)
{
    if (mesh) {
        SetSize(2*mesh->bsphere.r); 
        SetCenter(Vector3(mesh->bsphere.center[0], mesh->bsphere.center[1], 
            mesh->bsphere.center[2])); 
        feature = 0.5f * mesh->feature_size();
    } else errorf(("No trimesh setup"));
}

void
TriMeshWindow::
Keyboard(unsigned char c, int x, int y)
{
    switch (c)  {
        case 's':
            dflat = !dflat;
            break;
        case 'v':
            dvertices = !dvertices;
            break;
        case 'n':
            dnormals = !dnormals;
            break;
        default:
            ModelWindow::Keyboard(c, x, y);
            break;
    }
    GlutWMPostRedisplay();
}

TriMeshWindow::TriMeshWindow(void) 
{
    mesh = NULL;
    dflat = false; dvertices = false; dnormals = false;
}
